from typing import Optional, Any, List
from datetime import datetime

from pydantic import BaseModel


class Balance(BaseModel):
    status: str
    psys_cid: str
    currency: str
    balance: float


class TransactionsParams(BaseModel):
    order_number: int
    order_name: str
    source_currency: Optional[str] | None = None
    amount: Optional[float] | None = None
    source_rate: float


class TransactionsData(BaseModel):
    user_id: int
    shop_id: str
    type: str
    status: str
    pending_sum: float
    psys_cid: str
    currency: str
    source_currency: str
    source_rate: float
    fee: Optional[float] | None = None
    wallet_hash: str
    sendmany: Optional[Any] | None = None
    params: TransactionsParams
    expire_at_utc: datetime
    created_at_utc: datetime
    amount: float
    sum: float
    commission: float
    tx_url: List[str]
    tx_id: List[str]
    id: str
    actual_sum: float
    actual_commission: Optional[float] | None = None
    actual_fee: Optional[float] | None = None
    actual_invoice_sum: Optional[float] | None = None
    tx: Optional[str] | None = None
    confirmations: int
    status_code: int
    child_ids: List[str]
    parent_id: Optional[List[str]] | None = None



class Transactions(BaseModel):
    status: str
    data: TransactionsData



class FeePlanBaseData(BaseModel):
    key: Optional[str] | None = None
    name: Optional[str] | None = None
    description: Optional[str] | None = None
    gas_limit: Optional[str] | None = None
    gas_price: Optional[str] | None = None
    value: Optional[str] | None = None
    fee_unit: Optional[str] = None
    conf_target: Optional[str] | None = None
    fee_rate: Optional[int] | None = None
    fee_rate_unit: Optional[str] | None = None


class FeePlanEconomy(FeePlanBaseData):
    pass


class FeePlanNormal(FeePlanBaseData):
    pass


class FeePlanPriority(FeePlanBaseData):
    pass


class FeePlan(BaseModel):
    status: Optional[str] | None = None
    economy: Optional[FeePlanEconomy]  | None = None
    normal: Optional[FeePlanNormal] | None = None
    priority: Optional[FeePlanPriority] | None = None


class Comission(BaseModel):
    status: str
    commission: float
    fee: float
    plan: str
    use_wallet: bool
    use_wallet_balance: Optional[bool]
    plans: FeePlan


class InvoiceData(BaseModel):
    txn_id: str
    invoice_url: str
    amount: Optional[float] | None = None
    pending_amount: Optional[float] | None = None
    wallet_hash: Optional[str] | None = None
    psys_cid: Optional[str] | None = None
    currency: Optional[str] | None = None
    status: Optional[str] | None = None
    source_currency: Optional[str] | None = None
    source_rate: Optional[float] | None = None
    expire_utc: Optional[datetime] | None = None
    expected_confirmations: Optional[int] | None = None
    qr_code: Optional[str] | None = None
    verify_hash: Optional[str] | None = None
    invoice_commission: Optional[float] | None = None
    invoice_sum: Optional[float] | None = None
    invoice_total_sum: Optional[float] | None = None


class Invoice(BaseModel):
    status: str
    data: InvoiceData


from typing import Any
import aiohttp

from .plisio_exceptions import PlisioError


class _BaseAIOPlisioClient:
    _BASE_API_URL: str = "https://plisio.net/api/v1"

    def __init__(self, api_key: str) -> None:
        self._api_key = api_key

    async def __aenter__(self):
        return self
    
    async def __aexit__(self, exc_type, exc_value, traceback):
        return None
    
    async def _make_request(self, endpoint: str, post: bool=False, **kwargs) -> Any:
        async with aiohttp.ClientSession() as session:
            response = await session.get(f"{self._BASE_API_URL}/{endpoint}", **kwargs)
            if post:
                response = await session.post(f"{self._BASE_API_URL}/{endpoint}", **kwargs)
            data = await response.json()

            if "status" in data and data["status"] == "error":
                error_message, error_code = data["data"]["message"], data["data"]["code"]
                raise PlisioError(error_message, error_code)
            return data


from utils.base import _BaseAIOPlisioClient
from api.get_asyncio_api import _PlisioMethods, _PlisioInvoice

# TODO: Распределить код по файлам
# TODO: Написать датаклассы


class AIOPlisioClient(_BaseAIOPlisioClient):
    
    @property
    def get(self) -> _PlisioMethods:
        return _PlisioMethods(self._api_key)
    
    @property
    def invoice(self) -> _PlisioInvoice:
        return _PlisioInvoice(self._api_key)


from enum import Enum, EnumMeta, auto


class _EnumMeta(EnumMeta):
    def __getitem__(self, item):
        if item is not None:
            return super().__getitem__(item)
        return None


class CryptoCurrency(Enum, metaclass=_EnumMeta):
    ETH = "Ethereum"
    BTC = "Bitcoin"
    LTC = "Litecoin"
    DASH = "Dash"
    ZEC = "Zcash"
    DOGE = "Dogecoin"
    BCH = "Bitcoin Cash"
    XMR = "Monero"
    USDT = "Tether ERC-20"
    SHIB = "Shiba Inu",
    USDC = "USD Coin ERC-20",
    TRX = "Tron",
    USDT_TRX = "Tether TRC-20",
    BNB = "BNB Chain",
    USDT_BSC = "Teather BEP-20",
    BUSD = "Binance USD BEP-20",
    ETC = "Ethereum Classic",
    TZEC = "Zcash",
    BTT_TRX = "BitTorrent-Chain TRC-20",
    TUSD = "TrueUSD ERC-20"
    
    def __str__(self) -> str:
        return self.name


class FiatCurrency(Enum, metaclass=_EnumMeta):
    AED = "United Arab Emirates Dirham"
    AFN = "Afghan Afghani"
    ALL = "Albanian Lek"
    AMD = "Armenian Dram"
    ANG = "Netherlands Antillean Guilder"
    AOA = "Angolan Kwanza"
    ARS = "Argentine Peso"
    AUD = "Australian Dollar"
    AWG = "Aruban Florin"
    AZN = "Azerbaijani Manat"
    BAM = "Bosnia-Herzegovina Convertible Mark"
    BBD = "Barbadian Dollar"
    BDT = "Bangladeshi Taka"
    BGN = "Bulgarian Lev"
    BHD = "Bahraini Dinar"
    BIF = "Burundian Franc"
    BMD = "Bermuda Dollar"
    BND = "Brunei Dollar"
    BOB = "Bolivian Boliviano"
    BRL = "Brazilian Real"
    BSD = "Bahamian Dollar"
    BTN = "Bhutanese Ngultrum"
    BWP = "Botswana Pula"
    BYN = "New Belarusian Ruble"
    BYR = "Belarusian Ruble"
    BZD = "Belize Dollar"
    CAD = "Canadian Dollar"
    CDF = "Congolese Franc"
    CHF = "Swiss Franc"
    CLF = "Chilean Unit of Account (UF)"
    CLP = "Chilean Peso"
    CNY = "Chinese Yuan"
    COP = "Colombian Peso"
    CRC = "Costa Rican Colón"
    CUC = "Cuban Convertible Peso"
    CUP = "Cuban Peso"
    CVE = "Cape Verdean Escudo"
    CZK = "Czech Republic Koruna"
    DJF = "Djiboutian Franc"
    DKK = "Danish Krone"
    DOP = "Dominican Peso"
    DZD = "Algerian Dinar"
    EGP = "Egyptian Pound"
    ERN = "Eritrean Nakfa"
    ETB = "Ethiopian Birr"
    EUR = "Euro"
    FJD = "Fijian Dollar"
    FKP = "Falkland Islands Pound"
    GBP = "British Pound Sterling"
    GEL = "Georgian Lari"
    GGP = "Guernsey Pound"
    GHS = "Ghanaian Cedi"
    GIP = "Gibraltar Pound"
    GMD = "Gambian Dalasi"
    GNF = "Guinean Franc"
    GTQ = "Guatemalan Quetzal"
    GYD = "Guyanese Dollar"
    HKD = "Hong Kong Dollar"
    HNL = "Honduran Lempira"
    HRK = "Croatian Kuna"
    HTG = "Haitian Gourde"
    HUF = "Hungarian Forint"
    IDR = "Indonesian Rupiah"
    ILS = "Israeli New Sheqel"
    IMP = "Manx pound"
    INR = "Indian Rupee"
    IQD = "Iraqi Dinar"
    IRR = "Iranian Rial"
    ISK = "Icelandic Króna"
    JEP = "Jersey Pound"
    JMD = "Jamaican Dollar"
    JOD = "Jordanian Dinar"
    JPY = "Japanese Yen"
    KES = "Kenyan Shilling"
    KGS = "Kyrgyzstan Som"
    KHR = "Cambodian Riel"
    KMF = "Comorian Franc"
    KPW = "North Korean Won"
    KRW = "South Korean Won"
    KWD = "Kuwaiti Dinar"
    KYD = "Cayman Islands Dollar"
    KZT = "Kazakhstani Tenge"
    LAK = "Laotian Kip"
    LBP = "Lebanese Pound"
    LKR = "Sri Lankan Rupee"
    LRD = "Liberian Dollar"
    LSL = "Lesotho Loti"
    LTL = "Lithuanian Litas"
    LVL = "Latvian Lats"
    LYD = "Libyan Dinar"
    MAD = "Moroccan Dirham"
    MDL = "Moldovan Leu"
    MGA = "Malagasy Ariary"
    MKD = "Macedonian Denar"
    MMK = "Myanmar Kyat"
    MNT = "Mongolian Tugrik"
    MOP = "Macanese Pataca"
    MRO = "Mauritania Ouguiya"
    MUR = "Mauritian Rupee"
    MVR = "Maldivian Rufiyaa"
    MWK = "Malawi Kwacha"
    MXN = "Mexican Peso"
    MYR = "Malaysian Ringgit"
    MZN = "Mozambican Metical"
    NAD = "Namibian Dollar"
    NGN = "Nigerian Naira"
    NIO = "Nicaraguan Córdoba"
    NOK = "Norwegian Krone"
    NPR = "Nepalese Rupee"
    NZD = "New Zealand Dollar"
    OMR = "Omani Rial"
    PAB = "Panamanian Balboa"
    PEN = "Peruvian Nuevo Sol"
    PGK = "Papua New Guinean Kina"
    PHP = "Philippine Peso"
    PKR = "Pakistani Rupee"
    PLN = "Polish Zloty"
    PYG = "Paraguayan Guarani"
    QAR = "Qatari Rial"
    RON = "Romanian Leu"
    RSD = "Serbian Dinar"
    RUB = "Russian Ruble"
    RWF = "Rwandan Franc"
    SAR = "Saudi Riyal"
    SBD = "Solomon Islands Dollar"
    SCR = "Seychellois Rupee"
    SDG = "Sudanese Pound"
    SEK = "Swedish Krona"
    SGD = "Singapore Dollar"
    SHP = "Saint Helena Pound"
    SLL = "Sierra Leonean Leone"
    SOS = "Somali Shilling"
    SRD = "Surinamese Dollar"
    STD = "São Tomé and Príncipe Dobra"
    SVC = "Salvadoran Colón"
    SYP = "Syrian Pound"
    SZL = "Swazi Lilangeni"
    THB = "Thai Baht"
    TJS = "Tajikistan Somoni"
    TMT = "Turkmenistan Manat"
    TND = "Tunisian Dinar"
    TOP = "Tongan Paʻanga"
    TRY = "Turkish Lira"
    TTD = "Trinidad and Tobago Dollar"
    TWD = "New Taiwan Dollar"
    TZS = "Tanzanian Shilling"
    UAH = "Ukrainian Hryvnia"
    UGX = "Ugandan Shilling"
    USD = "United States Dollar"
    UYU = "Uruguayan Peso"
    UZS = "Uzbekistan Som"
    VEF = "Venezuelan Bolívar Fuerte"
    VND = "Vietnamese Dong"
    VUV = "Vanuatu Vatu"
    WST = "Samoan Tala"
    XAF = "CFA Franc BEAC"
    XAG = "Silver (troy ounce)"
    XAU = "Gold (troy ounce)"
    XCD = "East Caribbean Dollar"
    XDR = "Special Drawing Rights"
    XOF = "CFA Franc BCEAO"
    XPF = "CFP Franc"
    YER = "Yemeni Rial"
    ZAR = "South African Rand"
    ZMK = "Zambian Kwacha (pre-2013)"
    ZMW = "Zambian Kwacha"
    ZWL = "Zimbabwean Dollar"

    def __str__(self) -> str:
        return self.name


class TransactionStatus(Enum, metaclass=_EnumMeta):
    NEW = auto()
    PENDING = auto()
    EXPIRED = auto()
    COMPLETED = auto()
    MISMATCH = auto()
    ERROR = auto()
    CANCELLED = auto()

    def __str__(self) -> str:
        return self.name


class TransactionType(Enum, metaclass=_EnumMeta):
    CASH_IN = auto()
    CASH_OUT = auto()
    MASS_CASH_OUT = auto()
    INVOICE = auto()

    def __str__(self) -> str:
        return self.name


class PlanName(Enum, metaclass=_EnumMeta):
    ECONOMY = auto()
    NORMAL = auto()
    PRIORITY = auto()
    CUSTOM = auto()

    def __str__(self) -> str:
        return self.name


from setuptools import setup, find_packages
from io import open


def read(filename):
    with open(filename, "r", encoding="utf-8") as file:
        return file.read()


setup(
    name="aioplisio",
    version="1.0.0",
    description="Asynchronous wrapper for Plisio API",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="Fsoky",
    author_email="cyberuest0x12@gmail.com",
    url="https://github.com/Fsoky/aioplisio",
    keywords="api plisio crypto cryptopay plisio-api",
    packages=find_packages()
)


from typing import Optional, Any, List

from utils.base import _BaseAIOPlisioClient
from utils.enums import CryptoCurrency, FiatCurrency, TransactionStatus, TransactionType, PlanName
from utils.models import (
    Balance,
    Transactions,
    TransactionsData,
    TransactionsParams,
    FeePlan,
    FeePlanEconomy,
    FeePlanNormal,
    FeePlanPriority,
    Comission,
    Invoice,
    InvoiceData
)


class _PlisioMethods(_BaseAIOPlisioClient):

    async def balance(self, psys_cid: CryptoCurrency) -> Balance:
        payload = {
            "psys_cid": psys_cid,
            "api_key": self._api_key
        }

        response = await self._make_request("balances", params=payload)
        return Balance(status=response["status"], **response["data"])
    
    async def transactions(
        self,
        id: Optional[str] | None=None,
        *,
        page: Optional[int] | None=None,
        limit: Optional[int] | None=None,
        shop_id: Optional[str] | None=None,
        type: str | Optional[TransactionStatus] | None=None,
        status: str | Optional[TransactionStatus] | None=None,
        currency: str | Optional[CryptoCurrency] | None=None
    ) -> Transactions:
        payload = {
            "page": page,
            "limit": limit,
            "shop_id": shop_id,
            "type": type.name if not isinstance(type, str) and type is not None else type,
            "status": status.name if not isinstance(status, str) and type is not None else status,
            "currency": currency.name if not isinstance(currency, str) and type is not None else currency,
            "api_key": self._api_key
        }
        payload = {k: v for k, v in payload.items() if v is not None}
        
        endpoint = "operations" if id is None else f"operations/{id}"
        if endpoint.startswith("operations/"):
            payload = {"api_key": self._api_key}

        response = await self._make_request(endpoint, params=payload)
        response_data = response["data"].copy()
        response_data.pop("params", {})

        return Transactions(
            status=response["status"],
            data=TransactionsData(**response_data, params=TransactionsParams(**response["data"]["params"]))
        )
    
    async def fee_plan(self, psys_cid: str | CryptoCurrency) -> FeePlan:
        payload = {
            "api_key": self._api_key
        }

        response = await self._make_request(f"operations/fee-plan/{psys_cid}", params=payload)
        return FeePlan(
            status=response["status"],
            economy=FeePlanEconomy(**response["data"]["economy"]),
            normal=FeePlanNormal(**response["data"]["normal"]),
            priority=FeePlanPriority(**response["data"]["priority"])
        )
    
    async def fee(self, psys_cid: str | CryptoCurrency, addresses: str, amounts: float) -> Any:
        payload = {
            "addresses": addresses,
            "amounts": amounts,
            "api_key": self._api_key
        }

        response = await self._make_request(f"operations/fee/{psys_cid}", params=payload)
        return response
    
    async def commission(
        self,
        psys_cid: str | CryptoCurrency,
        addresses: str,
        amounts: float,
        fee_plan: PlanName,
        type: str | TransactionType
    ) -> Comission:
        payload = {
            "addresses": addresses,
            "amounts": amounts,
            "fee_plan": fee_plan,
            "type": type.name if not isinstance(type, str) else type,
            "api_key": self._api_key
        }

        response = await self._make_request(f"operations/commission/{psys_cid}", params=payload)
        return Comission(
            status=response["status"],
            commission=response["data"]["commission"],
            fee=response["data"]["fee"],
            plan=response["data"]["plan"],
            use_wallet=response["data"]["useWallet"],
            use_wallet_balance=response["data"]["useWalletBalance"],
            plans=FeePlan(
                normal=FeePlanNormal(**response["data"]["plans"]["normal"]),
                priority=FeePlanPriority(**response["data"]["plans"]["priority"])
            )
        )
    

class _PlisioInvoice(_BaseAIOPlisioClient):

    async def create(
        self,
        order_name: str,
        order_number: int,
        *,
        language: str="en_US",
        currency: Optional[str] | Optional[CryptoCurrency] | None=None,
        source_currency: Optional[str] | Optional[FiatCurrency] | None=None,
        source_amount: Optional[float] | None=None,
        amount: Optional[float] | None=None,
        allowed_psys_cids: Optional[List[str]] | Optional[List[CryptoCurrency]] | None=None,
        description: Optional[str] | None=None,
        callback_url: Optional[str] | None=None,
        success_callback_url: Optional[str] | None=None,
        fail_callback_url: Optional[str] | None=None,
        email: Optional[str] | None=None,
        plugin: Optional[Any] | None=None,
        version: Optional[Any] | None=None,
        redirect_to_invoice: Optional[str] | None=None,
        expire_min: Optional[int] | None=None
    ) -> Invoice:
        payload = {
            "order_name": order_name,
            "order_number": order_number,
            "language": language,
            "currency": currency.name if not isinstance(currency, str) else str(currency),
            "source_currency": source_currency.name if not isinstance(currency, str) else str(source_currency),
            "source_amount": source_amount,
            "amount": amount,
            "allowed_psys_cids": ",".join(
                    apc.name if not isinstance(apc, str) else apc
                    for apc in allowed_psys_cids
                ) if allowed_psys_cids is not None else None,
            "description": description,
            "callback_url": callback_url,
            "success_callback_url": success_callback_url,
            "fail_callback_url": fail_callback_url,
            "email": email,
            "plugin": plugin,
            "version": version,
            "redirect_to_invoice": redirect_to_invoice,
            "expire_min": expire_min,
            "api_key": self._api_key
        }
        payload = {k: v for k, v in payload.items() if v is not None}

        response = await self._make_request("invoices/new", params=payload)
        return Invoice(status=response["status"], data=InvoiceData(**response["data"]))